"""Lexer for Arrow."""
from typing import Optional, Union, Tuple, List
from .leaves import (
    Block, Body, Program, JumpUp, JumpDown, Function, End,
    Expression, Statement, Number, Boolean, Variable, String,
    Character, Array, BinaryOperator, UnaryOperator, BinaryOperation,
    UnaryOperation, Call, Index, Main, Operator, Assignment,
    PRECEDENCE, UNOPS, BINOPS, EQUALS, ARRREG, VarType, ArrayType,
    Input, Print, Declaration, Require
)

__all__ = ['ParsingFailed', 'ArrowParser']

class ParsingFailed(ValueError):
    """Parsing Arrow source failed."""
    def __init__(self, message, lineno, line):
        super().__init__(f'(text) line {lineno}: {line.strip()}\n{message}')
        self.lineno = lineno
        self.line = line

class ArrowParser:
    """Parse Arrow source."""

    source: Optional[str] = None

    def __init__(self, source: Optional[str] = None):
        """Initialize the parser, optionally with source."""
        self.source = source

    def parse(self, source: Optional[str] = None) -> Program:
        """Returns a Program representing the entire program.
        Raise ParsingFailed if parsing fails
        or TypeError if no text is given or set.
        """
        if source is None:
            if self.source is None:
                raise TypeError('No source in param or at init')
            source = self.source
        source = source.replace('\\\n', '')
        result = Program(body=Body(stmts=[]))
        stack = [result]
        mained = False
        try:
            for lineno, line in enumerate(source.splitlines()):
                statement = self.parse_line(line, len(stack))
                if statement is None:
                    continue
                if isinstance(statement, Block):
                    stack[-1].body.stmts.append(statement)
                    stack.append(statement)
                elif isinstance(statement, End):
                    if not isinstance(stack[-1], statement.cls):
                        raise AssertionError(
                            'Expecting end to {.__name__} '
                            'but got end to {.__name__} '
                            'instead'.format(
                                type(stack[-1]),
                                statement.cls
                            )
                        )
                    if statement.cls is JumpUp:
                        stack[-1].condition = statement.value
                    elif statement.cls is Function:
                        stack[-1].retval = statement.value
                    stack.pop()
                else:
                    stack[-1].body.stmts.append(statement)
                    if statement is Main():
                        assert not mained, 'Cannot have more than one main'
                        mained = True
                    statement.lineno = lineno + 1 # used in runner for errors
        except AssertionError as exc:
            raise ParsingFailed(exc.args[0], lineno + 1, line)
        return result

    def parse_line(self, line: str, stacklen: int) -> Union[
            None, Statement, Block, End
        ]:
        """Returns a single Statement,
        which can be None if the line was empty or a comment.
        """
        line = line.split('//', 1)[0].strip()
        if not line.lstrip('| \t'):
            return None # empty after removing comments
        # these are not affected by indents - they
        # must always be at the top level
        if line == 'function':
            # this line can be ignored for parsing purposes
            return None
        if line == 'main':
            return Main()
        try:
            # program               stacklen    start
            # /-->                  1           /-->
            # | /--< condition1     2           | /--<
            # | | /-->              3           | /-->
            # | | | do_thing        4           | do_thing
            # | | \--< condition2   4           \--<
            # | \-->                3           \-->
            # \--< condition3       2           \--<
            # conclusion: endings have no extra indents,
            # startings do, except on top level
            start = self.indent(0, line, stacklen)
            start, statement = self.statement(start, line, stacklen)
            return statement
        except IndexError:
            raise AssertionError('unexpected EOL')

    def whitespace(self, start: int, text: str, required: bool = False) -> int:
        """Skip past whitespace."""
        end = start
        try:
            while text[end].isspace():
                end += 1
        except IndexError:
            end = len(text)
        assert not (required and end == start), \
               f'column {start}: required whitespace not present'
        return end

    def indent(self, start: int, text: str, stacklen: int) -> int:
        """Skip past (and require) indentation."""
        # - 1 for root, - 1 for current level
        needed = max(stacklen - 2, 0)
        while needed > 0:
            assert text[start] in r'\|' or text[start].isspace(), \
                   f'Missing {needed} indent(s)'
            if text[start] == '|':
                needed -= 1
            start += 1
        start = self.whitespace(start, text)
        #if start == len(text):
        #    raise IndexError
        return start

    def statement(self, start: int, text: str,
                  stacklen: int) -> Tuple[int, Union[None, Statement,
                                                     Block, End]]:
        """Parse a statement."""
        arg = None
        if text.startswith(r'\--<', start):
            # end JumpUp and parse condition
            start = self.whitespace(start + 4, text)
            start, cond = self.expression(start, text)
            arg = End(cls=JumpUp, value=cond)
        elif text.startswith(r'\-->', start):
            # end JumpDown and None condition
            start += 4
            arg = End(cls=JumpDown, value=None)
        elif text.startswith('^', start):
            start = self.whitespace(start + 1, text)
            start, retval = self.expression(start, text)
            arg = End(cls=Function, value=retval)
        if arg is not None:
            return (start, arg)
        arg = None
        if stacklen > 1:
            start = self.indent(start, text, 3) # chop off last indent
        if text.startswith('/--<', start):
            # start JumpDown and parse condition
            start = self.whitespace(start + 4, text)
            start, cond = self.expression(start, text)
            arg = JumpDown(body=Body(stmts=[]), condition=cond)
        elif text.startswith('/-->', start):
            # two possibilities - either start JumpUp or Function
            start = self.whitespace(start + 4, text)
            try:
                start, typ = self.type(start, text)
            except AssertionError:
                # no type encountered, it's JumpUp
                arg = JumpUp(condition=None, body=Body(stmts=[]))
            else:
                start = self.whitespace(start, text, True)
                start, name = self.variable(start, text)
                start, args = self.args(start + 1, text)
                arg = Function(rettype=typ, name=name, params=args,
                               retval=None, body=Body(stmts=[]))
        elif text.startswith('print', start):
            start = self.whitespace(start + 5, text, True)
            start, arg = self.printstmt(start, text)
        elif text.startswith('require', start):
            start = self.whitespace(start + 7, text, True)
            start, arg = self.expression(start, text)
            arg = Require(expr=arg)
        if arg is not None:
            return (start, arg)
        try:
            start, arg = self.declaration(start, text)
        except AssertionError:
            pass # not a declaration
        else:
            return (start, arg)
        match = EQUALS.search(text, start)
        if match:
            start, name = self.expression(start, text[:match.start()])
            assert isinstance(name, Variable), \
                   f'Expected assignment target before column {start}'
            start = self.whitespace(match.end(), text)
            start, value = self.expression(start, text)
            arg = Assignment(name=name, value=value)
        else:
            start, arg = self.expression(start, text)
        return (start, arg)

    def printstmt(self, start: int, text: str) -> Print:
        """Parse a print statement."""
        end, args = self.call(start, text + ')')
        return (end - 1, Print(values=args))

    def expression(self, start: int, text: str, paren: bool = False) \
            -> Tuple[int, Optional[Expression]]:
        """Parse an expression."""
        if not text.strip():
            return (start, None)
        start, tokens = self.tokenize(start, text, paren)
        value = self.precedence(tokens)
        return (start, value)

    def precedence(self, tokens: List[Union[Expression, Operator]]) \
            -> Expression:
        """Decide precedence by converting infix to RPN.

        Uses the shunting-yard algorithm:
        https://en.wikipedia.org/wiki/Shunting-yard_algorithm
        """
        if len(tokens) == 1:
            assert isinstance(tokens[0], Expression), 'Unexpected lone operator'
            # skip all the stack manipulation
            return tokens[0]
        output = []
        stack = []
        while tokens:
            token = tokens.pop(0)
            if isinstance(token, Expression):
                output.append(token)
            elif isinstance(token, Operator):
                try:
                    while (isinstance(stack[-1], Operator)
                           and (PRECEDENCE[stack[-1]] >= PRECEDENCE[token])):
                        output.append(stack.pop())
                except IndexError:
                    pass # stack is empty, no need to pop anything
                stack.append(token)
        while stack:
            # transfer rest of stack to output
            output.append(stack.pop())
        # construct (a op b)s from RPN
        stack.clear()
        for item in output:
            if isinstance(item, Expression):
                stack.append(item)
            elif isinstance(item, BinaryOperator):
                arg2 = stack.pop()
                arg1 = stack.pop()
                stack.append(BinaryOperation(arg1=arg1, arg2=arg2,
                                             operator=item))
            elif isinstance(item, UnaryOperator):
                stack.append(UnaryOperation(arg=stack.pop(), operator=item))
        assert len(stack) == 1, 'Something has gone wrong - report as bug'
        return stack[0]

    def tokenize(self, start: int, text: str, paren: bool = False) \
            -> Tuple[int, List[Union[Expression, Operator]]]:
        """Split expression string into tokenized expressions/operators."""
        #breakpoint()
        tokens = []
        while start < len(text):
            while start < len(text):
                # test unary operators, add them if valid
                match = UNOPS.match(text, start)
                try:
                    if match:
                        tokens.append(UnaryOperator(match.group(0)))
                        start += len(match.group(0))
                    else:
                        break
                finally:
                    start = self.whitespace(start, text)
            # parse another arbitrary expression
            # (which could be a parenthesized full expr)
            start, arg = self.expression_token(start, text, paren)
            tokens.append(arg)
            start = self.whitespace(start, text)
            if start >= len(text):
                break # no more tokens
            # test binary operators, add one if valid
            match = BINOPS.match(text, start)
            if match:
                tokens.append(BinaryOperator(match.group(0)))
                start += len(match.group(0))
            elif text[start] in ')]},':
                break # expressions end here no matter what
            else:
                raise AssertionError(f'Expected operator at column {start}')
            start = self.whitespace(start, text)
        assert tokens and isinstance(tokens[-1], Expression), \
               f'Expected expression at column {start}'
        return (start, tokens)

    def expression_token(self, start: int, text: str, paren: bool = False) \
            -> Tuple[int, Expression]:
        """Expression with no operators, for use as a token."""
        if text.startswith('true', start):
            # boolean value
            start += 4
            arg1 = Boolean(value=True)
        elif text.startswith('false', start):
            start += 5
            arg1 = Boolean(value=False)
        elif text.startswith('input', start):
            start += 5
            start = self.whitespace(start, text, True)
            if text.startswith('int', start):
                arg = VarType.INT
            elif text.startswith('char[]', start):
                arg = ArrayType(type=VarType.CHAR)
            else:
                raise AssertionError('input type must be int or char[]')
            start += len(str(arg))
            arg1 = Input(type=arg)
        elif text[start].isdecimal():
            # number value
            start, arg1 = self.number(start, text)
        elif text[start].isidentifier():
            # variable
            start, arg1 = self.variable(start, text)
            if start < len(text):
                if text[start] == '(':
                    # function call
                    start, args = self.call(start + 1, text)
                    assert text[start] == ')', f'Missing ) at column {start}'
                    start += 1
                    start = self.whitespace(start, text)
                    arg1 = Call(name=arg1, args=args)
                while start < len(text) and text[start] == '[':
                    # array indexation
                    start, index = self.expression(start + 1, text, True)
                    assert text[start] == ']', f'Missing ] at column {start}'
                    start += 1
                    start = self.whitespace(start, text)
                    arg1 = Index(name=arg1, index=index)
        elif text[start] == "'":
            # character
            start, arg1 = self.character(start + 1, text)
        elif text[start] == '"':
            # string
            start, arg1 = self.string(start + 1, text)
        elif text[start] == '{':
            # array
            start, arg1 = self.array(start + 1, text)
        elif text[start] == '(':
            # recursion!
            start, arg1 = self.expression(start + 1, text, True)
        else:
            raise AssertionError(f'Expected expression at column {start}')
        if start < len(text) and paren and text[start] == ')':
            start += 1
        return (start, arg1)

    def number(self, start: int, text: str) -> Tuple[int, Number]:
        """Parse a number."""
        end = start
        while end < len(text) and text[end].isdecimal():
            end += 1
        return (end, Number(value=int(text[start:end])))

    def variable(self, start: int, text: str) -> Tuple[int, Variable]:
        """Parse a variable (or function) name."""
        end = start + 1
        while end < len(text) and (
                text[end].isidentifier()
                or text[end].isdecimal()
        ):
            end += 1
        assert text[start:end].isidentifier(), \
               f'Invalid identifier at column {start}'
        return (end, Variable(name=text[start:end]))

    def character(self, start: int, text: str) -> Tuple[int, Character]:
        """Parse a character literal."""
        end = start
        while text[end] != "'":
            if text[end] == '\\':
                end += 1
                if text[end] in r"\'":
                    end += 1
            else:
                end += 1
        char = text[start:end].encode().decode('unicode-escape')
        assert len(char) == 1, f'Invalid character literal at column {start}'
        return (end + 1, Character(value=char))

    def string(self, start: int, text: str) -> Tuple[int, String]:
        """Parse a string literal."""
        end = start
        while text[end] != '"':
            if text[end] == '\\':
                end += 1
                if text[end] in r'\"':
                    end += 1
            else:
                end += 1
        stri = text[start:end].encode().decode('unicode-escape')
        return (end + 1, String(value=stri))

    def array(self, start: int, text: str) -> Tuple[int, Array]:
        """Parse an array literal."""
        end = start
        values = []
        while text[end] != '}':
            end = self.whitespace(end, text)
            end, value = self.expression(end, text)
            values.append(value)
            end = self.whitespace(end, text)
            if text[end] == ',':
                end += 1
        return (end + 1, Array(values=values))

    def call(self, start: int, text: str) -> Tuple[int, List[Expression]]:
        """Parse arguments of a function call."""
        end = start
        args = []
        while end < len(text) and text[end] != ')':
            end = self.whitespace(end, text)
            end, value = self.expression(end, text)
            args.append(value)
            end = self.whitespace(end, text)
            if end < len(text) and text[end] == ',':
                end += 1
        return (end, args)

    def args(self, start: int, text: str) -> Tuple[int, List[Declaration]]:
        """Parse arguments of a function definition."""
        end = start
        args = []
        while end < len(text) and text[end] != ')':
            end = self.whitespace(end, text)
            end, value = self.declaration(end, text, True)
            args.append(value)
            end = self.whitespace(end, text)
            if end < len(text) and text[end] == ',':
                end += 1
        return (end + 1, args)

    def declaration(self, start: int, text: str, param: bool = False) \
            -> Tuple[int, Declaration]:
        """Parse a variable/parameter declaration."""
        end, typ = self.type(start, text)
        end = self.whitespace(end, text, True)
        if param:
            end, expr = self.variable(end, text)
        else:
            end, expr = self.expression(end, text)
            assert isinstance(expr, Variable), \
                   f'Invalid declaration at column {start}'
        return (end, Declaration(type=typ, name=expr))

    def type(self, start: int, text: str) \
            -> Tuple[int, VarType]:
        """Parse a type."""
        for item in VarType:
            # pylint: disable=no-member
            if text.startswith(str(item), start):
                arg = item
                start += len(str(item))
                while ARRREG.match(text, start):
                    start = self.whitespace(start, text)
                    arg = ArrayType(type=arg)
                    start += 2
                break
            # pylint: enable=no-member
        else:
            raise AssertionError(f'Expected type at column {start}')
        return (start, arg)



"""For running `python -m arrowlang`."""
import sys
import argparse
from arrowlang import run

def main(args=None):
    """The entry point of the module."""
    argparser = argparse.ArgumentParser()
    argparser.add_argument('file', nargs='?', help='A file to run. '
                           'Use - for stdin without prompt.')
    args = argparser.parse_args(args)
    if args.file:
        if args.file == '-':
            source = sys.stdin.read()
        else:
            with open(args.file) as sfile:
                source = sfile.read()
    else:
        print('Enter Arrow program. Type ^D (Linux) or ^Z then Enter (Windows)'
              ' to finish.')
        source = sys.stdin.read()
    run(source)

if __name__ == '__main__':
    main()



"""Leaves of the parse tree."""
from __future__ import annotations
import re
import json # just for string representation
from enum import Enum
from typing import Optional, Union, List, Type, get_type_hints

class MemEnum(Enum):
    """Adds a method to enums to check if a value is an enum value."""

    @classmethod
    def has(cls, value: str) -> bool:
        """Returns True if the value of an enum == value."""
        values = set(item.value for item in cls)
        return value in values

    def __str__(self):
        return self.value

class Operator(MemEnum):
    """Operators ABC."""
    pass

class BinaryOperator(Operator):
    """Binary operators."""
    EQUAL = '=='
    INEQUAL = '!='
    LESS = '<'
    MORE = '>'
    PLUS = '+'
    MINUS = '-'
    TIMES = '*'
    DIVIDE = '/'
    MODULO = '%'
    AND = 'and'
    OR = 'or'

BINOPS = re.compile(r'(==|!=|<|>|\+|-|\*|\/|%|and\b|or\b)')
EQUALS = re.compile(r'(?<![=!])=(?!=)')

class UnaryOperator(Operator):
    """Unary operators."""
    NEG = '-'
    POS = '+'
    NOT = 'not'
    LEN = 'length'

UNOPS = re.compile(r'(\+|-|not\b|length\b)')

ARRREG = re.compile(r'\s*\[\]')

PRECEDENCE = {
    # These are arbitrary numbers. They are only compared.
    # The spacing of 10 is in case more operators are added.
    UnaryOperator.LEN: 80,

    UnaryOperator.POS: 70,
    UnaryOperator.NEG: 70,

    BinaryOperator.TIMES: 60,
    BinaryOperator.DIVIDE: 60,
    BinaryOperator.MODULO: 60,

    BinaryOperator.PLUS: 50,
    BinaryOperator.MINUS: 50,

    BinaryOperator.LESS: 40,
    BinaryOperator.MORE: 40,
    BinaryOperator.INEQUAL: 40,
    BinaryOperator.EQUAL: 40,

    UnaryOperator.NOT: 30,
    BinaryOperator.AND: 20,
    BinaryOperator.OR: 10,
}

class VarType(MemEnum):
    """Variable types."""
    BOOL = 'bool'
    INT = 'int'
    CHAR = 'char'

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

BOOLOPS = {BinaryOperator.AND, BinaryOperator.OR, UnaryOperator.NOT}
INTOPS = {BinaryOperator.LESS, BinaryOperator.MORE, BinaryOperator.PLUS,
          BinaryOperator.MINUS, BinaryOperator.TIMES, BinaryOperator.DIVIDE,
          BinaryOperator.MODULO, UnaryOperator.NEG, UnaryOperator.POS}
SAMEOPS = {BinaryOperator.EQUAL, BinaryOperator.INEQUAL}
ARROPS = {UnaryOperator.LEN}

RETURNOPS = {
    VarType.BOOL: {
        BinaryOperator.EQUAL, BinaryOperator.INEQUAL, BinaryOperator.LESS,
        BinaryOperator.MORE, BinaryOperator.AND, BinaryOperator.OR,
        UnaryOperator.NOT
    },
    VarType.INT: {
        BinaryOperator.PLUS, BinaryOperator.MINUS, BinaryOperator.TIMES,
        BinaryOperator.DIVIDE, BinaryOperator.MODULO, UnaryOperator.NEG,
        UnaryOperator.POS, UnaryOperator.LEN
    }
}
OPRETURNS = {_op: _ret for _ret, _ops in RETURNOPS.items() for _op in _ops}

def _dataclass(cls: type) -> type:
    """Decorator to add kwarg-only __init__"""
    def typingcheck(value, typ):
        try:
            #if typ is Any:
            #    return True
            if getattr(typ, '__origin__', None) is Union:
                typ = typ.__args__
            if getattr(typ, '__origin__', None) is type:
                return issubclass(value, typ.__args__)
            return isinstance(value, typ)
        except TypeError:
            print(value, typ)
            raise
    def __init__(self, **kwargs):
        cls = type(self)
        cls.__types__ = get_type_hints(cls)
        for var in cls.__types__:
            if cls.__types__[var] is None:
                cls.__types__[var] = type(None)
            elif getattr(cls.__types__[var], '__origin__', None) is list:
                cls.__types__[var] = list
        for var, typ in self.__types__.items():
            if var not in kwargs and not hasattr(cls, var):
                raise TypeError('__init__() missing required '
                                f'keyword argument: {var!r}')
            elif var not in kwargs:
                kwargs[var] = getattr(cls, var)
            if not typingcheck(kwargs[var], typ):
                if isinstance(typ, tuple):
                    typ = ', '.join(f'{i.__name__!r}' for i in typ)
                else:
                    typ = f'{typ!r}'
                raise TypeError(f'{var!r} is not {typ}: {kwargs[var]!r}')
            setattr(self, var, kwargs[var])
    def __repr__(self):
        ret = type(self).__name__
        ret += '('
        ret += ', '.join(
            f'{k!s}={[...] if isinstance(v, list) else v!r}'
            for k, v in self.__dict__.items()
            if not k.startswith('__')
            and self.__types__.get(k, None) is not None
        )
        ret += ')'
        return ret
    def __setattr__(self, key, value):
        if key.startswith('__'):
            super(cls, self).__setattr__(key, value)
            return
        typ = self.__types__[key]
        if typ in (None, type(None)):
            super(cls, self).__setattr__(key, value)
            return
        if getattr(typ, '__origin__', None) is Union:
            typ = self.__types__[key] = typ.__args__
        if not typingcheck(value, typ):
            raise TypeError(f'{key!r} is not {typ.__name__!r}: {value!r}')
        super(cls, self).__setattr__(key, value)
    cls.__init__ = __init__
    cls.__repr__ = __repr__
    cls.__setattr__ = __setattr__
    cls.__hash__ = lambda self: hash(str(self))
    cls.__eq__ = lambda self, other: hash(self) == hash(other)
    return cls

@_dataclass
class ArrayType:
    """Array types."""
    # None means empty
    type: Union[None, VarType, 'ArrayType']

    def __str__(self):
        return str(self.type) + '[]'

# ABCs

@_dataclass
class Statement:
    """ABC for a single statement."""
    lineno: Optional[int] = None

class Expression(Statement):
    """ABC for expressions."""
    pass

# Possible kinds of expressions

class Variable(Expression):
    """Variables are expressions"""
    name: str

    def __str__(self):
        return self.name

class Number(Expression):
    """Numbers are expressions"""
    value: int

    def __str__(self):
        return str(self.value)

class Boolean(Expression):
    """Booleans are expressions"""
    value: bool

    def __str__(self):
        return str(self.value).lower()

class Character(Expression):
    """Character literals"""
    value: str

    def __str__(self):
        return repr(self.value)

class String(Expression):
    """String literals"""
    value: str

    def __str__(self):
        return json.dumps(self.value)

class Array(Expression):
    """Array literals"""
    values: List[Expression]

    def __str__(self):
        return '{' + ', '.join(map(str, self.values)) + '}'

class Operation(Expression):
    """Operation ABC."""
    operator: Operator

class BinaryOperation(Operation):
    """Binary operator"""
    arg1: Expression
    arg2: Expression
    operator: BinaryOperator

    def __str__(self):
        return f'({self.arg1!s} {self.operator.value} {self.arg2!s})'

class UnaryOperation(Operation):
    """Unary operator"""
    arg: Expression
    operator: UnaryOperator

    def __str__(self):
        return f'({self.operator.value}{" " if len(self.operator.value) != 1 else ""}{self.arg!s})'

class Call(Expression):
    """Function call"""
    name: Variable
    args: List[Expression]

    def __str__(self):
        return f'{self.name!s}({", ".join(map(str, self.args))})'

# subclasses Variable because it's
# an assignment target too
class Index(Variable):
    """Array indexing"""
    name: Expression
    index: Expression

    def __str__(self):
        return f'{self.name!s}[{self.index!s}]'

class Input(Expression):
    """Getting input from user"""
    type: Union[VarType, ArrayType]

    def __str__(self):
        return f'(input {self.type.value})'

# Possible kinds of statements

class Declaration(Statement):
    """Variable type declaration"""
    type: Union[VarType, ArrayType]
    name: Variable

    def __str__(self):
        return f'{self.type!s} {self.name!s}'

class Assignment(Statement):
    """Assigning variable value"""
    name: Variable
    value: Expression

    def __str__(self):
        return f'{self.name!s} = {self.value!s}'

class Require(Statement):
    """Strict assert"""
    expr: Expression

    def __str__(self):
        return f'require {self.expr!s}'

class Print(Statement):
    """Output"""
    values: List[Expression]

    def __str__(self):
        return 'print ' + ', '.join(map(str, self.values))

class Main(Statement):
    """Start of the program."""
    _instance = None
    def __new__(cls):
        """There is only one ``main`` statement."""
        if cls._instance is None:
            cls._instance = Statement.__new__(cls)
        return cls._instance

    def __str__(self):
        return 'main'

# Statements come together in blocks

@_dataclass
class Block:
    """Multiple statements (or blocks!)"""

    def _indent(self, body):
        return '\n'.join('| ' + line for line in str(body).splitlines())

class Body(Block):
    """Body of a block"""
    stmts: List[Union[Block, Statement]]

    def __str__(self):
        return '\n'.join(map(str, self.stmts))

class JumpDown(Block):
    r"""
    /--< condition
    | block
    \-->
    """
    condition: Optional[Expression]
    body: Body

    @property
    def lineno(self) -> int:
        """Line number of JumpDown is line number of its condition."""
        return self.body.stmts[0].lineno - 1

    def __str__(self):
        body = self._indent(self.body)
        return f'/--< {self.condition!s}\n{body!s}\n\\-->'

class JumpUp(Block):
    r"""
    /-->
    | block
    \--< condition
    """
    condition: Optional[Expression]
    body: Body

    @property
    def lineno(self) -> int:
        """Line number of JumpUp is line number of its condition."""
        return self.body.stmts[-1].lineno + 1

    def __str__(self):
        body = self._indent(self.body)
        return f'/-->\n{body!s}\n\\--> {self.condition!s}'

class Function(Block):
    """
    /--> type funcname(type params)
    | block
    ^ retval
    """
    rettype: Union[VarType, ArrayType]
    name: Variable
    params: List[Declaration] # abuse this type
    retval: Optional[Expression]
    body: Body

    @property
    def lineno(self) -> int:
        """Line number of Function is line number of the function keyword."""
        return self.body.stmts[0].lineno - 2

    def __str__(self):
        body = self._indent(self.body)
        return f'function\n/--> {self.rettype!s} {self.name!s}' \
               f'({", ".join(map(str, self.params))})' \
               f'\n{body!s}\n^ {self.retval!s}'

@_dataclass
class Program:
    """The root."""
    body: Body

    def __str__(self):
        return str(self.body)

@_dataclass
class End:
    """Marks the end of a block."""
    cls: Type[Block]
    value: Optional[Expression]



"""
An implementation of Arrow[1] in Python.

[1]: https://github.com/jacob-g/arrow-lang
"""
from __future__ import annotations
from .parser import ArrowParser
from .runner import ArrowRunner

__version__ = '0.0.0'
__all__ = ['run', 'ArrowRunner', 'ArrowParser']

def run(source: str, **kwargs) -> None:
    """Run Arrow ``source``."""
    ArrowRunner(source).run(**kwargs)



"""Typechecker and runner for Arrow."""
from typing import Optional, Union, Mapping, Tuple, Type, Any
from .leaves import (
    Program, Main, Function, Declaration, VarType, ArrayType, JumpDown,
    BOOLOPS, INTOPS, SAMEOPS, ARROPS, OPRETURNS, Assignment, JumpUp, Block,
    Require, Print, Expression, Index, Number, Boolean, Character, String,
    Variable, Array, Operation, Call, BinaryOperation, UnaryOperation, Input,
)
from .parser import ArrowParser

class TypecheckFailed(TypeError):
    """Typechecking failed."""
    def __init__(self, message, lineno, line):
        super().__init__(f'(text) line {lineno}: {line.strip()}\n{message}')
        self.lineno = lineno
        self.line = line

class ArrowError(RuntimeError):
    """Typechecking failed."""
    def __init__(self, message, lineno, line):
        super().__init__(f'(text) line {lineno}: {line.strip()}\n{message}')
        self.lineno = lineno
        self.line = line

class ArrowTyping:
    """Typecheck Arrow code."""

    source: Optional[Program] = None

    # same name can be reused in function,
    # so record types per function (and main)
    vartypes: Mapping[Union[Main, Function],
                      Mapping[Variable, VarType]]
    # map function name to function obj
    functions: Mapping[Variable, Function]

    def __init__(self, source: Optional[Program] = None):
        """Initialize the typechecker, optionally with source."""
        self.source = source
        self.vartypes = {Main(): {}}
        self.functions = {}

    def check(self, source: Optional[Program] = None) -> Tuple[
            Mapping[Union[Main, Function],
                    Mapping[Variable, VarType]],
            Mapping[Variable, Function]
        ]:
        """Typechecks the Arrow source.
        Raises TypeError if no text is given or set,
        or TypecheckFailed (which subclasses TypeError!)
        if typechecking found an error.
        """
        if source is None:
            if self.source is None:
                raise TypeError('No source in param or at init')
            source = self.source
        self.check_stmts(source)
        return (self.vartypes, self.functions)

    def check_stmts(self, block: Union[Program, Block],
                    func: Union[Main, Function] = Main()):
        """Recursively typecheck a block."""
        try:
            for stmt in block.body.stmts:
                if isinstance(stmt, Function):
                    assert not isinstance(block, Function), \
                           'Cannot have nested functions'
                    self.functions[stmt.name] = stmt
                    self.vartypes[stmt] = {}
                    # declare variables
                    for param in stmt.params:
                        self.set_vartype(stmt, param)
                    self.check_stmts(stmt, stmt)
                    self.check_expr(stmt, stmt.retval, stmt.rettype)
                elif isinstance(stmt, Block):
                    self.check_expr(func, stmt.condition, VarType.BOOL)
                    # recursively check that block
                    self.check_stmts(stmt, func)
                elif isinstance(stmt, Declaration):
                    # type varname
                    self.set_vartype(func, stmt)
                elif isinstance(stmt, Assignment):
                    # var = value
                    self.check_expr(func, stmt.value,
                                    self.get_vartype(func, stmt.name))
                elif isinstance(stmt, Require):
                    self.check_expr(func, stmt.expr, VarType.BOOL)
                elif isinstance(stmt, Print):
                    # print expr, expr, expr
                    for expr in stmt.values:
                        self.check_expr(func, expr, None)
                elif isinstance(stmt, Main):
                    pass
                else:
                    self.check_expr(func, stmt, None)
        except AssertionError as exc:
            sstmt = str(stmt)
            if isinstance(stmt, (JumpDown, Function)):
                sstmt = sstmt.splitlines()[0]
            elif isinstance(stmt, JumpUp):
                sstmt = sstmt.splitlines()[-1]
            raise TypecheckFailed(exc.args[0], stmt.lineno, sstmt)

    def set_vartype(self, func: Union[Main, Function],
                    decl: Declaration):
        """Declare a variable's type.
        Raises AssertionError if already declared.
        """
        # setdefault 1 adds the new block to the mapping,
        # setdefault 2 will only set [decl.name] = decl
        # if decl.name was not already in the mapping.
        # if it was, the value will not is decl
        # and we raise hell. Otherwise it's already set.
        name = decl.name
        names = 0
        while isinstance(name, Index):
            name = name.name
            names += 1
        typ = decl.type
        types = 0
        while isinstance(typ, ArrayType):
            typ = typ.type
            types += 1
        assert types >= names, f'{name!s} has too many dimensions declared ' \
               f'({names} > {types})'
        assert self.vartypes.setdefault(func, {})\
               .setdefault(name, decl) is decl, \
               f'{name!s} is already declared at line ' \
               f'{self.vartypes[func][name].lineno}'

    def get_vartype(self, func: Union[Main, Function],
                    name: Variable):
        """Return the type of a variable.
        Raises AssertionError if not declared.
        """
        if isinstance(name, Index):
            return self.get_vartype(func, name.name).type
        assert name in self.vartypes[func] or name in self.vartypes[Main()], \
               f'unknown variable {name!s}'
        return self.vartypes[func][name].type

    def check_expr(self, func: Union[Main, Function], expr: Expression,
                   typ: Union[None, type(...), VarType, Type[ArrayType]]=None) \
                   -> Union[None, VarType, ArrayType]:
        """Recursively check an expression's type.
        None means the result of the expression need not be typechecked
        but its components still do need it.
        """
        inst = typ is ArrayType
        if isinstance(expr, Index):
            self.check_expr(func, expr.index, VarType.INT)
            self.check_expr(func, expr.name, ArrayType)
        if isinstance(expr, Variable) and typ is not None:
            vartype = self.get_vartype(func, expr)
            if typ is ...:
                return vartype
            if inst:
                assert isinstance(vartype, typ), \
                       f'{expr!s} is {vartype!s}, not array'
            else:
                assert typ == vartype, \
                       f'{expr!s} is {vartype!s}, not {typ!s}'
        elif isinstance(expr, Number) and typ is not None:
            if typ is ...:
                return VarType.INT
            if inst:
                typ = 'array'
            assert typ == VarType.INT, f'{expr!s} is int, not {typ!s}'
        elif isinstance(expr, Boolean) and typ is not None:
            if typ is ...:
                return VarType.BOOL
            if inst:
                typ = 'array'
            assert typ == VarType.BOOL, f'{expr!s} is bool, not {typ!s}'
        elif isinstance(expr, Character) and typ is not None:
            if typ is ...:
                return VarType.CHAR
            if inst:
                typ = 'array'
            assert typ == VarType.CHAR, f'{expr!s} is char, not {typ!s}'
        elif isinstance(expr, String) and typ is not None:
            if typ is ...:
                return ArrayType(type=VarType.CHAR)
            if not inst: # strings are arrays
                assert typ == ArrayType(type=VarType.CHAR), \
                       f'{expr!s} is char[], not {typ!s}'
        elif isinstance(expr, Array) and typ is not None:
            if expr.values:
                itype = self.check_expr(func, expr.values[0], ...)
                for val in expr.values:
                    try:
                        self.check_expr(func, val, itype)
                    except AssertionError:
                        raise AssertionError('array must be '
                                             'all one type') from None
            if typ is ...:
                return ArrayType(type=itype) \
                       if expr.values \
                       else ArrayType(type=None)
            if not inst:
                assert typ == ArrayType(type=itype) \
                       if expr.values \
                       else isinstance(typ, ArrayType), \
                       f'{expr!s} is an array, not {typ!s}'
        elif isinstance(expr, Operation):
            opertype = self.check_oper(func, expr)
            if typ is ...:
                return opertype
            if typ is not None:
                if inst:
                    assert isinstance(opertype, typ), f'{expr!s} is ' \
                           f'{opertype!s}, not array'
                else:
                    assert typ == opertype, f'{expr!s} is {opertype!s}, ' \
                           f'not {typ!s}'
        elif isinstance(expr, Call):
            assert expr.name in self.functions, 'unknown function {expr.name!s}'
            called = self.functions[expr.name]
            if typ is ...:
                return called.rettype
            if typ is not None:
                if inst:
                    assert isinstance(called.rettype, typ), \
                           f'{called.name!s}() returns {called.rettype!s}, ' \
                           'not array'
                else:
                    assert typ == called.rettype, \
                           f'{called.name!s}() returns {called.rettype!s}, ' \
                           f'not {typ!s}'
            assert len(expr.args) == len(called.params), \
                   f'{called.name!s}() expects {len(called.params)} ' \
                   f'arguments, not {len(expr.args)}'
            for arg, param in zip(expr.args, called.params):
                self.check_expr(func, arg, param.type)
        elif isinstance(expr, Input) and typ is not None:
            if typ is ...:
                return expr.type
            if inst:
                assert isinstance(expr.type, typ), 'expected array but ' \
                       f'got {expr.type!s}'
            else:
                assert typ == expr.type, f'expected {typ!s} but got ' \
                       f'{expr.type!s}'
        elif typ is None:
            return None
        else:
            raise TypeError('How did you get here? Report this as a bug.')
        return None

    def check_oper(self, func: Union[Main, Function],
                   expr: Operation) -> VarType:
        """Check an operation's arguments depending on the operation."""
        if isinstance(expr, UnaryOperation):
            if expr.operator in INTOPS:
                self.check_expr(func, expr.arg, VarType.INT)
            elif expr.operator in BOOLOPS:
                self.check_expr(func, expr.arg, VarType.BOOL)
            elif expr.operator in ARROPS:
                try:
                    self.check_expr(func, expr.arg, ArrayType)
                except AssertionError:
                    raise AssertionError('length x expected an array') from None
            else:
                raise TypeError('How did you get here? Report this as a bug.')
            return OPRETURNS[expr.operator]
        if expr.operator in INTOPS:
            self.check_expr(func, expr.arg1, VarType.INT)
            self.check_expr(func, expr.arg2, VarType.INT)
        elif expr.operator in BOOLOPS:
            self.check_expr(func, expr.arg1, VarType.BOOL)
            self.check_expr(func, expr.arg2, VarType.BOOL)
        elif expr.operator in SAMEOPS:
            typ1 = self.check_expr(func, expr.arg1, ...)
            typ2 = self.check_expr(func, expr.arg2, ...)
            assert typ1 == typ2, f'Operands to {expr.operator!s} ' \
                   'must be of same type'
        else:
            raise TypeError('How did you get here? Report this as a bug.')
        return OPRETURNS[expr.operator]

class ArrowRunner:
    """Run Arrow code."""

    source: Union[Program, str, None] = None
    # these two attributes come from those in ArrowTyping
    vartypes: Mapping[Union[Main, Function],
                      Mapping[Variable, VarType]]
    functions: Mapping[Variable, Function]
    # these are specific
    variables: Mapping[Union[Main, Function],
                       Mapping[Variable, Any]]

    def __init__(self, source: Union[Program, str, None] = None):
        """Initialize the runner, optionally with source."""
        self.source = source

    def run(self, source: Union[Program, str, None] = None) -> None:
        """Runs the Arrow source.
        Raises TypeError if no text is given or set.
        """
        if source is None:
            if self.source is None:
                raise TypeError('No source in param or at init')
            source = self.source
        else:
            self.source = source
        if isinstance(source, str):
            self.source = ArrowParser(self.source).parse()
        self.vartypes, self.functions = ArrowTyping(self.source).check()
        self.variables = {}
        self.run_func(self.source, Main())

    def run_func(self, block: Union[Program, Block],
                 func: Union[Main, Function]):
        """Call a function."""
        # a key thing to note here is that it's expected that
        # the typechecker has already been run.
        # This means there is a lot less validation as to
        # whether the variable should or shouldn't exist, for
        # example - the typechecker should have caught that.
        # For example, the Assignment branch assumes a Declaration
        # has already been encountered, since for it to have passed
        # the typechecker it in fact has to have been.
        try:
            for stmt in block.body.stmts:
                if isinstance(stmt, Function) and isinstance(func, Main):
                    self.variables[stmt] = {}
                    continue
                if stmt is Main():
                    continue
                if isinstance(stmt, JumpDown):
                    if not self.run_expr(stmt.condition, func):
                        self.run_func(stmt, func)
                elif isinstance(stmt, JumpUp):
                    self.run_func(stmt, func)
                    while self.run_expr(stmt.condition, func):
                        self.run_func(stmt, func)
                elif isinstance(stmt, Declaration):
                    self.variables.setdefault(func, {})
                    typ = stmt.type
                    while isinstance(typ, ArrayType):
                        typ = typ.type
                    ddims = []
                    name = stmt.name
                    while isinstance(name, Index):
                        ddims.append(name.index)
                        name = name.name
                    def copy(value: Union[list, None]) -> Union[list, None]:
                        """Shallow copy a list, do nothing to None."""
                        if value is None:
                            return None
                        return value[:]
                    bottom = None
                    while ddims:
                        bottom = [copy(bottom)
                                  for _ in range(self.run_expr(ddims.pop(),
                                                               func))]
                    self.variables.setdefault(func, {})[name] = bottom
                elif isinstance(stmt, Assignment):
                    name = stmt.name
                    values = []
                    while isinstance(name, Index):
                        values.append(name.index)
                        name = name.name
                    if values:
                        var = self.variables[func][name]
                        for idx in values[:-1]:
                            var = var[self.run_expr(idx, func)]
                        var[self.run_expr(values[-1], func)] \
                            = self.run_expr(stmt.value, func)
                    else:
                        self.variables[func][name] = self.run_expr(stmt.value,
                                                                   func)
                elif isinstance(stmt, Require):
                    if not self.run_expr(stmt.expr, func):
                        # don't use assert because require is *strict* assert
                        raise AssertionError('require statement failed')
                elif isinstance(stmt, Print):
                    values = []
                    for val in stmt.values:
                        val = self.run_expr(val, func)
                        if isinstance(val, list) and (not val or isinstance(
                                val[0], str)):
                            val = ''.join(val)
                        values.append(val)
                    print(*values)
        except Exception as exc:
            sstmt = str(stmt)
            if isinstance(stmt, (JumpDown, Function)):
                sstmt = sstmt.splitlines()[0]
            elif isinstance(stmt, JumpUp):
                sstmt = sstmt.splitlines()[-1]
            raise ArrowError(exc.args[0], stmt.lineno, sstmt)

    def run_expr(self, expr: Expression, func: Union[Main, Function]) -> Any:
        """Evaluate and return the semantic value of an expression."""
        if isinstance(expr, Variable):
            name = expr
            idxs = []
            while isinstance(name, Index):
                idxs.append(name.index)
                name = name.name
            var = self.variables[func][name]
            for idx in idxs:
                assert var is not None, 'using uninitialized array'
                var = var[self.run_expr(idx, func)]
            return var
        if isinstance(expr, (Number, Boolean, Character)):
            return expr.value
        if isinstance(expr, String):
            return list(expr.value)
        if isinstance(expr, Array):
            return [self.run_expr(i, func) for i in expr.values]
        if isinstance(expr, BinaryOperation):
            return self.run_binop(expr, func)
        if isinstance(expr, UnaryOperation):
            return self.run_unop(expr, func)
        if isinstance(expr, Call):
            return self.call(expr, func)
        if isinstance(expr, Input):
            inp = input()
            if expr.type == VarType.INT:
                return int(inp)
            if expr.type == ArrayType(type=VarType.CHAR):
                return list(inp)
            raise TypeError('How did you get here? Report this as a bug.')
        raise TypeError('How did you get here? Report this as a bug.')

    def run_binop(self, expr: BinaryOperation,
                  func: Union[Main, Function]) -> Any:
        """Evaluate and return the semantic value of a binary operation."""
        # deal with short circuiting first
        oper = str(expr.operator)
        if oper == 'and':
            arg1 = self.run_expr(expr.arg1, func)
            if arg1:
                return self.run_expr(expr.arg2, func)
            return False
        if oper == 'or':
            arg1 = self.run_expr(expr.arg1, func)
            if arg1:
                return arg1
            return self.run_expr(expr.arg2, func)
        # this is fine, because if any of the previous branches
        # match, they ultimately return, so the expressions
        # are never evaluated twice
        arg1 = self.run_expr(expr.arg1, func)
        arg2 = self.run_expr(expr.arg2, func)
        if oper == '+':
            return arg1 + arg2
        if oper == '-':
            return arg1 - arg2
        if oper == '*':
            return arg1 * arg2
        if oper == '/':
            return arg1 // arg2 # arrow division is integer division
        if oper == '%':
            return arg1 % arg2
        if oper == '==':
            return arg1 == arg2
        if oper == '!=':
            return arg1 != arg2
        if oper == '<':
            return arg1 < arg2
        if oper == '>':
            return arg1 > arg2
        raise TypeError('How did you get here? Report this as a bug.')

    def run_unop(self, expr: UnaryOperation,
                 func: Union[Main, Function]) -> Any:
        """Evaluate and return the semantic value of a unary operation."""
        arg = self.run_expr(expr.arg, func)
        oper = str(expr.operator)
        # pylint doesn't seem to realize that arg can be int
        # pylint: disable=invalid-unary-operand-type
        if oper == '-':
            return -arg
        if oper == '+':
            return +arg
        if oper == 'not':
            return not arg
        if oper == 'length':
            return len(arg)
        # pylint: enable=invalid-unary-operand-type
        raise TypeError('How did you get here? Report this as a bug.')

    def call(self, expr: Call, func: Union[Main, Function]) -> Any:
        """Call a function."""
        called = self.functions[expr.name]
        for param, value in zip(called.params, expr.args):
            name = param.name
            while isinstance(name, Index):
                name = name.name
            self.variables[called][name] = self.run_expr(value, func)
        self.run_func(called, called)
        return self.run_expr(called.retval, called)



from setuptools import setup

with open('README.rst') as f:
    long_description = f.read()

setup(name='py-arrow-lang',
      version='2020.06.18',
      description='An implementation of Arrow ' \
      '(https://github.com/jacob-g/arrow-lang) in Python.',
      long_description=long_description,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development :: Interpreters',
      ],
      keywords='arrow language interpreter',
      url='https://github.com/Kenny2github/py-arrow-lang',
      author='Kenny2github',
      author_email='kenny2minecraft@gmail.com',
      license='MIT',
      packages=['arrowlang'],
      entry_points={
          'console_scripts': [
              'arrow=arrowlang.__main__:main',
          ],
      },
)
